import UIKit

// What is a PROTOCOL?
//
// A protocol is a CONTRACT between two or more classes. it's declared as a set
// of FUNCTION STUBS. I fa class SUBSCRIBES to a protocol, that means it
// PROMISES to implement the functions that the protocol requires. Example:


// This protocol is a CONTRACT that basically says
// "You must implement a function called print which returns a String."
protocol Printable {
    func print() -> String
}

// A class can SUBSCRIBE to this protocol by following the class name
// with a colon and the name of the protocol you want to subscribe to. For
// example (NOTE: you'll have to uncomment the following code in order to
//                see the code breaking--highlight the code block and hit
//                Cmd+/ to toggle comments.)

//class BankAccount : Printable {
//    let accountName: String
//    let accountBalance: Double
//    
//    init(accountName: String, accountBalance: Double) {
//        self.accountName = accountName
//        self.accountBalance = accountBalance
//    }
//}

// Notice that Xcode has a problem with this otherwise-perfectly-fine class.
// The error reads:
//      "Type 'BankAccount' does not conform to protocol 'Printable'."
// This is because BankAccount is subscribed to Printable, which is a contract.
// The BankAccount isn't fulfilling its end of the contract. So, if we
// re-define BankAccount to implement a method called print(), it'll work:

class BankAccount : Printable {
    let accountName: String
    let accountBalance: Double

    init(accountName: String, accountBalance: Double) {
        self.accountName = accountName
        self.accountBalance = accountBalance
    }
    
    func print() -> String {
        return "BankAccount(\(self.accountName), \(self.accountBalance))"
    }
}

// Notice that it doesn't matter HOW a function is implemented, just that
// it is implemented. So, our implementation of print() could've easily been
//
// func print() -> String {
//     return "Screw off."
// }
//
// Now, you can treat a protocol as a class, of sorts. What I mean by this
// is this is a valid chunk of code:

var someObject: Printable?
if (someObject != nil) {
    println(someObject!.print())
}

// So even though Printable isn't actually a class, we can still use the
// functions associated with it as if it were a class. Now, the above code
// won't do anything because someObject is defaulted to nil, but we can try a
// more concrete example:

var somePrintableObject: Printable = BankAccount(accountName: "Mike",
                                                 accountBalance: 0)
println(somePrintableObject.print())

// So, we can treat BankAccount as an object of type Printable. Where can this
// be particularly useful? Well, here's an example. Let's first define a 
// different class which can also subscribe to the Printable protocol:

class StudentLoanAccount : Printable {
    let accountName: String
    let amountOfDebt: Double
    
    init(accountName: String, amountOfDebt: Double) {
        self.accountName = accountName
        self.amountOfDebt = amountOfDebt
    }
    
    func print() -> String {
        return "StudentLoanAccount(\(self.accountName), \(self.amountOfDebt))"
    }
}

// Notice the above is SIMILAR to the BankAccount class, but not the same.
// They print slightly different things.

let bankAccount = BankAccount(accountName: "Mike", accountBalance: 0)
let studentLoanAccount = StudentLoanAccount(accountName: "Mike",
                                            amountOfDebt: 9999999)
let accounts : [Printable] = [bankAccount, studentLoanAccount]

// See, we just grouped two different classes into one common list of Printable
// objects. Now:

for account in accounts {
    println(account.print())
}

// This prints:
//
// BankAccount(Mike, 0.0)
// StudentLoanAccount(Mike, 9,999,999.00)
//
// So, as we can see, protocols are a nice way to group together objects that
// share similar funcitonality. As expected, functions can take protocol-
// objects as input and output, so we can see something like:

func allCapsDescription(printable: Printable) {
    println(printable.print().uppercaseString)
}

allCapsDescription(studentLoanAccount)

// So for most intents and purposes, protocols can act as classes.

// Protocols can have more than one function in its "contract". So, we could
// have, for example:

protocol Numerical {
    func convertToDouble() -> Double
    func convertToFloat() -> Float
    func convertToUInt() -> UInt
    func convertToInt() -> Int
}

// And of course, a class which subscribes to that protocol is contractually
// obligated to implement ALL of these methods, like so

class Integer : Numerical {
    let number: Int
    
    init(number: Int) {
        self.number = number
    }
    
    func convertToDouble() -> Double {
        return Double(self.number)
    }
    
    func convertToFloat() -> Float {
        return Float(self.number)
    }
    
    func convertToUInt() -> UInt {
        return UInt(self.number)
    }
    
    func convertToInt() -> Int {
        return self.number
    }
}

// But we notice that the last function, convertToInt, isn't actually
// necessary. Similar, if we were to make classes FloatingPoint,
// DoublePrecisionFloat, or UnsignedInteger, the corresponding Float,
// Double, and UInt functions would be useless. So, one solution to overzealous
// protocols like this is to make some functions optional.
//
// Objective-C ahs this feature, but Swift doesn't. Luckily, the connections
// between Swift and Objective-C run EXTREMELY deep, and the two languages
// very frequently intermix.
//
// We first have to turn a protocol into an Objective-C protocol first. We do
// this by prepending @objc to the 'protocol <name>' line. Even though it's an
// Objective-C protocol, it functions exactly the same as a Swift protocol,
// just with a different name.
//
// Next, in order to make a function in a protocol optional, simply prepend
// the line with 'optional'. So, a newer, more optionalized version of our
// Numerical protocol above is as follows:

@objc protocol Numerical2 {
    optional func convertToDouble() -> Double
    optional func convertToFloat() -> Float
    optional func convertToUInt() -> UInt
    optional func convertToInt() -> Int
}

// Now, we could say:

class FloatingPoint : Numerical2 {
    let number: Float
    
    init (number: Float) {
        self.number = number
    }
    
    func convertToDouble() -> Double {
        return Double(self.number)
    }
    
    func convertToUInt() -> UInt {
        return UInt(self.number)
    }
    
    func convertToInt() -> Int {
        return Int(self.number)
    }
}

// And now Xcode doesn't yell at us for not implementing the convertToFloat
// function. However, what happens if we try to call convertToFloat on
// the FloatingPoint class? The Numerical2 protocol says it *can* implement
// that method, not this it *does*. Well, the answer to that question depends
// on how you type an object like FloatingPoint. Suppose you make an object
// of the FloatingPoint like so:

let float1: FloatingPoint = FloatingPoint(number: 1.0)

// If you try to run the following command:

//float1.convertToFloat()   // Error: uncomment to run.

// The error reads
//  "'FloatingPoint' does not have a member named 'convertToFloat'"
// So, this is what we'd expect. However, if we make the following version
// of a FloatingPoint object:

let float2: Numerical2 = FloatingPoint(number: 2.0)

// If we try the command:

//float2.convertToFloat()  // Error: uncomment to run.

// The error reads
//   "Value of optional type '(()->Float)?' not unwrapped; ..."
//
// So, the function is considered an Optional! In a functional language like
// Swift, functions are first-order objects like Int and String and whatnot, so
// they can totally be Optionals. Because they're an Optional type, they have to
// be unwrapped, like so:

float2.convertToFloat!()

// And the above line of code is valid! Now, it's going to cause an error
// because the convertToFloat function doesn't exist, so we're effectively
// unwrapping nil.

// So, when we treat an object's type as that of a protocol, any optional
// functions become Function Optionals. So, if we want to call a different
// method on float2, we'd have to unwrap it first:

float2.convertToInt!()

// While float1 wouldn't require such.

float1.convertToInt()

// So, while classes and protocols definitely aren't the same, it's definitely
// worth knowing the inherent differences and advantages.
//
// In general, you won't have to write protocols with optional functions, but
// you definitely are going to subscribe to protocols with optional functions.
//
// You can do a lot of other obscure things with protocols, but we won't cover
// them today.

